@@ -1,9 +1,7 @@
 import pandas as pd
 import ray
 from collections import deque
-from pyquokka.dataset import RedisObjectsDataset
 import pickle
-import redis
 from threading import Lock
 import time
 import gc
@@ -11,11 +9,13 @@ import sys
 import polars
 import pyarrow as pa
 import types
-import concurrent.futures
+import duckdb
+import psutil
 import pyarrow.flight
 # isolated simplified test bench for different fault tolerance protocols
 
 VERBOSE = False
+PROFILE = False
 
 # above this limit we are going to start flushing things to disk
 INPUT_MAILBOX_SIZE_LIMIT = 1024 * 1024 * 1024 * 2 # you can have 2GB in your input mailbox
@@ -23,8 +23,13 @@ INPUT_MAILBOX_SIZE_LIMIT = 1024 * 1024 * 1024 * 2 # you can have 2GB in your inp
 def convert_to_format(batch, format):
     format = pickle.loads(format.to_pybytes())
     if format == "polars":
+
+        aligned_batch = pa.record_batch([pa.concat_arrays([arr]) for arr in batch], schema=batch.schema)
+        return polars.from_arrow(pa.Table.from_batches([aligned_batch]))
+
         # ideally you use zero copy, but this could result in bugs sometimes. https://issues.apache.org/jira/browse/ARROW-17783
-        return polars.from_pandas(pa.Table.from_batches([batch]).to_pandas())
+        # return polars.from_pandas(pa.Table.from_batches([batch]).to_pandas())
+        # return polars.from_arrow(pa.Table.from_batches([batch]))
     elif format == "pandas":
         return batch.to_pandas()
     elif format == "custom":
@@ -77,13 +82,7 @@ class Node:
         self.channel = channel
 
         self.targets = {}
-        self.r = redis.Redis(host='localhost', port=6800, db=0)
-        self.head_r = redis.Redis(host=ray.get_runtime_context().gcs_address.split(":")[0], port=6800, db=0)
-        #self.plasma_client = plasma.connect("/tmp/plasma")
-
-        self.target_rs = {}
-        self.target_ps = {}
-
+        
         self.flight_clients = {}
         self.flight_client = pyarrow.flight.connect("grpc://0.0.0.0:5005")
 
@@ -102,55 +101,21 @@ class Node:
         self.out_seq[node_id] = {channel: 0 for channel in channel_to_ip}
 
         unique_ips = set(channel_to_ip.values())
-        redis_clients = {i: redis.Redis(host=i, port=6800, db=0) if i != self.ip else redis.Redis(host='localhost', port = 6800, db=0) for i in unique_ips}
         self.targets[node_id] = (channel_to_ip, target_info)
-        self.target_rs[node_id] = {}
-        self.target_ps[node_id] = {}
         
         self.flight_clients[node_id] = {}
 
         flight_clients = {i: pyarrow.flight.connect("grpc://" + str(i) + ":5005") if i != self.ip else pyarrow.flight.connect("grpc://0.0.0.0:5005") for i in unique_ips}
         for channel in channel_to_ip:
             self.flight_clients[node_id][channel] = flight_clients[channel_to_ip[channel]]
-
-        for channel in channel_to_ip:
-            self.target_rs[node_id][channel] = redis_clients[channel_to_ip[channel]]
         
-        for client in redis_clients:
-            pubsub = redis_clients[client].pubsub(ignore_subscribe_messages = True)
-            pubsub.subscribe("node-done-"+str(node_id))
-            self.target_ps[node_id][channel] = pubsub
         
         self.alive_targets[node_id] = {i for i in channel_to_ip}
         # remember the self.strikes stuff? Now we cannot check for that because a downstream target could just die.
         # it's ok if we send stuff to dead people. Fault tolerance is supposed to take care of this.
-        
-    def update_targets(self):
-
-        for target_node in self.target_ps:
-            # there are #-ip locations you need to poll here.
-            for channel in self.target_ps[target_node]:
-                client = self.target_ps[target_node][channel]
-                while True:
-                    message = client.get_message()
-                    
-                    if message is not None:
-                        if VERBOSE:
-                            print(message['data'])
-                        self.alive_targets[target_node].remove(int(message['data']))
-                        if len(self.alive_targets[target_node]) == 0:
-                            self.alive_targets.pop(target_node)
-                    else:
-                        break 
-        if len(self.alive_targets) > 0:
-            return True
-        else:
-            return False
 
     def push(self, data):
             
-        
-
         '''
         Quokka should have some support for custom data types, similar to DaFt by Eventual.
         Since we transmit data through Arrow Flight, we need to convert the data into arrow record batches.
@@ -172,14 +137,21 @@ class Node:
             my_format = "custom"
             print("custom data does not support predicates and projections")
             raise NotImplementedError
+        
+        # if type(data) == pd.core.frame.DataFrame:
+        #     data = pa.Table.from_pandas(data)
+        # elif type(data) == polars.internals.DataFrame:
+        #     data = data.to_arrow()
+        # elif type(data) == pa.lib.Table:
+        #     pass
+        # elif type(data) == pa.lib.RecordBatch:
+        #     data = pa.Table.from_batches([data])
+        # else:
+        #     my_format = "custom"
+        #     print("custom data does not support predicates and projections")
+        #     raise NotImplementedError
 
         # downstream targets are done. You should be done too then.
-        try:    
-            if not self.update_targets():
-                return False
-        except:
-            if VERBOSE:
-                print("downstream failure detected")
 
         for target in self.alive_targets:
             original_channel_to_ip, target_info = self.targets[target]
@@ -191,14 +163,23 @@ class Node:
 
             assert target_info.lowered
 
-            #print("pre-filer",data)
+            start = time.time()
+            if target_info.predicate != True:
+                data = data.filter(target_info.predicate(data))
+            # # data = self.con.execute("select * from data where " + target_info.predicate).arrow()
+            #data = polars.from_arrow(data)
+            # #print("post-filer", data)
+
+            if PROFILE:
+                print("filter ", time.time() - start)
             
-            data = data.filter(target_info.predicate(data))
-            #print("post-filer", data)
-
+            start = time.time()
             partitioned_payload = target_info.partitioner.func(data, self.channel, len(original_channel_to_ip))
-            #print("post-partition", partitioned_payload)
 
+            if PROFILE:
+                print("partition ", time.time() - start)
+
+            
             assert type(partitioned_payload) == dict and max(partitioned_payload.keys()) < len(original_channel_to_ip)
             
             for channel in self.alive_targets[target]:
@@ -210,11 +191,17 @@ class Node:
                     payload = partitioned_payload[channel]
 
                     # the batch funcs here all expect arrow data fomat.
+
+                    start = time.time()
+
                     for func in target_info.batch_funcs:
                         if len(payload) == 0:
                             payload = None
                             break
                         payload = func(payload)
+                    
+                    if PROFILE:
+                        print("batch func time per channel ", time.time() - start)
 
                 if payload is None or len(payload) == 0:
                     payload = None
@@ -229,6 +216,8 @@ class Node:
                 
                 client = self.flight_clients[target][channel]
                 batches, my_format = convert_from_format(payload)
+
+                start = time.time()
                 
                 # format (target id, target channel, source id, source channel, tag, format)
                 #print("pushing to target",target,"channel",channel,"from source",self.id,"channel",self.channel,"tag",self.out_seq)
@@ -240,7 +229,7 @@ class Node:
                     result = next(client.do_action(action))
                     #print(result.body.to_pybytes().decode("utf-8"))
                     if result.body.to_pybytes().decode("utf-8") != "True":
-                        time.sleep(1)
+                        time.sleep(0.1)
                     else:
                         break
 
@@ -249,6 +238,9 @@ class Node:
                 for batch in batches:
                     writer.write_batch(batch)
                 writer.close()
+            
+                if PROFILE:
+                    print("push time per channel ", time.time() - start)
        
         return True
 
@@ -257,14 +249,6 @@ class Node:
         if VERBOSE:
             print("IM DONE", self.id, self.channel, time.time())
 
-        try:    
-            if not self.update_targets():
-                if VERBOSE:
-                    print("WIERD STUFF IS HAPPENING")
-                return False
-        except:
-            print("downstream failure detected")
-
         for target in self.alive_targets:
             for channel in self.alive_targets[target]:
                 self.out_seq[target][channel] += 1
@@ -295,10 +279,16 @@ class InputNode(Node):
         
     def execute(self):
 
+        pa.set_cpu_count(8)
+        #self.con = duckdb.connect()
+        #self.con.execute('PRAGMA threads=%d' % 8)
         self.input_generator = self.accessor.get_next_batch(self.channel, None)
         for pos, batch in self.input_generator:
             if batch is not None and len(batch) > 0:
+                start = time.time()
                 self.push(batch)
+                if PROFILE:
+                    print("pushing", time.time() - start)
 
         if VERBOSE:
             print("INPUT DONE", self.id, self.channel, time.time())
@@ -381,7 +371,9 @@ class NonBlockingTaskNode(TaskNode):
         while True:
 
             # be nice. 
-            time.sleep(0.01)
+            time.sleep(0.001)
+            pa.set_cpu_count(8)
+            # self.con = duckdb.connect()
             batch_info, should_terminate = self.get_batches()
             if should_terminate:
                 break
@@ -455,46 +447,65 @@ class BlockingTaskNode(TaskNode):
     
     def execute(self):
 
+        pa.set_cpu_count(8)
+        spill_no = 0
+        # self.con = duckdb.connect()
+
         add_tasks = []
 
         while True:
+
             batch_info, should_terminate = self.get_batches()
             if should_terminate:
                 break
             # deque messages from the mailbox in a way that makes sense
             
             stream_id, requests = self.schedule_for_execution(batch_info)
+            
+            start = time.time()
+
             if stream_id is None:
                 continue
             request = ((self.id, self.channel), requests)
 
             reader = self.flight_client.do_get(pyarrow.flight.Ticket(pickle.dumps(request)))
 
+            #print("making reader", time.time() - start)
+            start = time.time()
+
             batches = []
             while True:
                 try:
+                    #start = time.time()
                     chunk, metadata = reader.read_chunk()
+                    #print("read time", time.time() - start)
+                    #start = time.time()
                     batches.append(convert_to_format(chunk, metadata))
+                    #print("convert time", time.time() - start)
                 except StopIteration:
                     break
             
-            
+            #print("reading batches", time.time() - start)
+            start = time.time()
+
             batches = [i for i in batches if i is not None]
             results = self.functionObject.execute( batches,self.physical_to_logical_mapping[stream_id], self.channel)
+
+            #print("executing", time.time() - start)
             
             if results is not None and len(results) > 0:
                 cursor = 0
                 stride = 1000000
                 while cursor < len(results):
-                    key = str(self.id) + "-" + str(self.channel) + "-" + str(self.object_count)
-                    self.object_count += 1
-                    try:
-                        self.r.set(key, pickle.dumps(results[cursor : cursor + stride]))
-                    except:
-                        print(results)
-                        raise Exception
-                    # we really should be doing sys.getsizeof(result), but that doesn't work for polars dfs
-                    add_tasks.append(self.output_dataset.added_object.remote(self.channel, (ray.util.get_node_ip_address(), key, stride)))
+
+                    if psutil.virtual_memory().percent < 70:
+                        add_tasks.append(self.output_dataset.added_object.remote(self.channel, [ray.put(results[cursor: cursor + stride].to_arrow(), _owner = self.output_dataset)]))
+                    else:
+                        filename = "/data/spill-" + str(self.id) + "-" + str(self.channel) + "-" + str(spill_no) + ".parquet"
+                        results[cursor: cursor + stride].write_parquet(filename)
+                        add_tasks.append(self.output_dataset.added_parquet.remote(self.channel, filename))
+                        spill_no += 1
+
                     cursor += stride
             else:
                 pass
@@ -504,13 +515,14 @@ class BlockingTaskNode(TaskNode):
         if type(obj_done) == types.GeneratorType:
             for object in obj_done:
                 if object is not None:
-                    key = str(self.id) + "-" + str(self.channel) + "-" + str(self.object_count)
-                    self.object_count += 1
-                    self.r.set(key, pickle.dumps(object))
-                    if hasattr(object, "__len__"):
-                        add_tasks.append(self.output_dataset.added_object.remote(self.channel, (ray.util.get_node_ip_address(), key, len(object))))
+                    if psutil.virtual_memory().percent < 70:
+                        add_tasks.append(self.output_dataset.added_object.remote(self.channel, [ray.put(object.to_arrow(), _owner = self.output_dataset)]))
                     else:
-                        add_tasks.append(self.output_dataset.added_object.remote(self.channel, (ray.util.get_node_ip_address(), key,sys.getsizeof(object))))
+                        filename = "/data/spill-" + str(self.id) + "-" + str(self.channel) + "-" + str(spill_no) + ".parquet"
+                        object.write_parquet(filename)
+                        add_tasks.append(self.output_dataset.added_parquet.remote(self.channel, filename))
+                        spill_no += 1
+
             del self.functionObject
             gc.collect()
         else:
@@ -518,14 +530,13 @@ class BlockingTaskNode(TaskNode):
             gc.collect()
             
             if obj_done is not None:
-                key = str(self.id) + "-" + str(self.channel) + "-" + str(self.object_count)
-                self.object_count += 1
-                self.r.set(key, pickle.dumps(obj_done))
-                if hasattr(obj_done, "__len__"):
-                    add_tasks.append(self.output_dataset.added_object.remote(self.channel, (ray.util.get_node_ip_address(), key, len(obj_done))))
+                if psutil.virtual_memory().percent < 70:
+                    add_tasks.append(self.output_dataset.added_object.remote(self.channel, [ray.put(obj_done.to_arrow(), _owner = self.output_dataset)]))
                 else:
-                    add_tasks.append(self.output_dataset.added_object.remote(self.channel, (ray.util.get_node_ip_address(), key, sys.getsizeof(obj_done))))
-     
+                    filename = "/data/spill-" + str(self.id) + "-" + str(self.channel) + "-" + str(spill_no) + ".parquet"
+                    obj_done.write_parquet(filename)
+                    add_tasks.append(self.output_dataset.added_parquet.remote(self.channel, filename))
+                    spill_no += 1
         ray.get(add_tasks)
         ray.get(self.output_dataset.done_channel.remote(self.channel))
         
